// Copyright (C) 2018-2024 Intel Corporation
// SPDX-License-Identifier: Apache-2.0
//

#include "zero_executor.hpp"

#include <ze_api.h>

#include <functional>
#include <iostream>
#include <sstream>
#include <string>

#include "intel_npu/al/config/common.hpp"
#include "intel_npu/al/itt.hpp"
#include "intel_npu/al/prefix.hpp"
#include "openvino/runtime/properties.hpp"
#include "ze_command_queue_npu_ext.h"
#include "zero_device.hpp"
#include "zero_utils.hpp"

using namespace intel_npu;

ZeroExecutor::ZeroExecutor(const std::shared_ptr<const ZeroInitStructsHolder>& initStructs,
                           const std::shared_ptr<const NetworkDescription>& networkDescription,
                           const Config& config,
                           const uint32_t& group_ordinal)
    : _config(config),
      _logger("Graph", _config.get<LOG_LEVEL>()),
      _initStructs(initStructs),
      _networkDesc(networkDescription),
      _graph_ddi_table_ext(_initStructs->getGraphDdiTable()),
      _group_ordinal(group_ordinal),
      _command_queues{{std::make_shared<CommandQueue>(_initStructs->getDevice(),
                                                      _initStructs->getContext(),
                                                      zeroUtils::toZeQueuePriority(_config.get<MODEL_PRIORITY>()),
                                                      _initStructs->getCommandQueueDdiTable(),
                                                      _config,
                                                      group_ordinal),
                       std::make_shared<CommandQueue>(_initStructs->getDevice(),
                                                      _initStructs->getContext(),
                                                      zeroUtils::toZeQueuePriority(_config.get<MODEL_PRIORITY>()),
                                                      _initStructs->getCommandQueueDdiTable(),
                                                      _config,
                                                      group_ordinal),
                       std::make_shared<CommandQueue>(_initStructs->getDevice(),
                                                      _initStructs->getContext(),
                                                      zeroUtils::toZeQueuePriority(_config.get<MODEL_PRIORITY>()),
                                                      _initStructs->getCommandQueueDdiTable(),
                                                      _config,
                                                      group_ordinal)}} {
    _logger.debug("create graph_command_list");
    OV_ITT_SCOPED_TASK(itt::domains::LevelZeroBackend, "Executor::ZeroExecutor");
    CommandList graph_command_list(_initStructs->getDevice(),
                                   _initStructs->getContext(),
                                   _initStructs->getGraphDdiTable(),
                                   _config,
                                   _group_ordinal);
    _logger.debug("create graph_command_queue");
    CommandQueue graph_command_queue(_initStructs->getDevice(),
                                     _initStructs->getContext(),
                                     ZE_COMMAND_QUEUE_PRIORITY_NORMAL,
                                     _initStructs->getCommandQueueDdiTable(),
                                     _config,
                                     _group_ordinal);
    _logger.debug("create fence");
    Fence fence(graph_command_queue, _config);

    OV_ITT_TASK_CHAIN(ZERO_EXECUTOR_GRAPH, itt::domains::LevelZeroBackend, "Executor::ZeroExecutor", "graphCreate");

    // _graph is a nullptr for CIP path, a new handle will be obtained from the driver based on the given
    // compiledNetwork _graph gets (reuses) graphHandle from the compiler for CID path
    if (_networkDesc->metadata.graphHandle == nullptr) {
        _logger.debug("create graph handle on executor");
        ze_graph_desc_t desc{ZE_STRUCTURE_TYPE_GRAPH_DESC_PROPERTIES,
                             nullptr,
                             ZE_GRAPH_FORMAT_NATIVE,
                             _networkDesc->compiledNetwork.size(),
                             _networkDesc->compiledNetwork.data(),
                             nullptr};

        zeroUtils::throwOnFail(
            "pfnCreate",
            _graph_ddi_table_ext->pfnCreate(_initStructs->getContext(), _initStructs->getDevice(), &desc, &_graph));

    } else {
        _logger.debug("reuse graph handle created from compiler");
        _graph = static_cast<ze_graph_handle_t>(_networkDesc->metadata.graphHandle);
    }

    OV_ITT_TASK_NEXT(ZERO_EXECUTOR_GRAPH, "pfnGetProperties");
    _logger.debug("performing pfnGetProperties");
    zeroUtils::throwOnFail("pfnGetProperties", _graph_ddi_table_ext->pfnGetProperties(_graph, &_props));
    auto targetDriverExtVersion = _initStructs->getDriverExtVersion();
    if (targetDriverExtVersion <= ZE_GRAPH_EXT_VERSION_1_1) {
        OPENVINO_THROW("Incompatibility between the NPU plugin and driver! The driver version is too old, please "
                       "update the driver version");
    }

    OV_ITT_TASK_NEXT(ZERO_EXECUTOR_GRAPH, "pfnGetArgumentProperties3");
    _logger.debug("performing pfnGetArgumentProperties3");
    for (uint32_t index = 0; index < _props.numGraphArgs; ++index) {
        ze_graph_argument_properties_3_t arg3;
        zeroUtils::throwOnFail("pfnGetArgumentProperties3",
                               _graph_ddi_table_ext->pfnGetArgumentProperties3(_graph, index, &arg3));

        if (arg3.type == ZE_GRAPH_ARGUMENT_TYPE_INPUT) {
            _input_descriptors.push_back(ArgumentDescriptor{arg3, index});
        } else {
            _output_descriptors.push_back(ArgumentDescriptor{arg3, index});
        }
    }

    OV_ITT_TASK_NEXT(ZERO_EXECUTOR_GRAPH, "appendGraphInitialize");
    _logger.debug("performing appendGraphInitialize");
    graph_command_list.appendGraphInitialize(_graph);
    _logger.debug("closing graph command list");
    graph_command_list.close();

    OV_ITT_TASK_NEXT(ZERO_EXECUTOR_GRAPH, "queue_execute");
    _logger.debug("performing executeCommandList");
    graph_command_queue.executeCommandList(graph_command_list, fence);
    _logger.debug("performing hostSynchronize");
    fence.hostSynchronize();
    _logger.debug("hostSynchronize completed");

    if (config.has<WORKLOAD_TYPE>()) {
        setWorkloadType(config.get<WORKLOAD_TYPE>());
    }
}

void ZeroExecutor::setWorkloadType(const ov::WorkloadType workloadType) const {
    ze_command_queue_workload_type_t zeWorkloadType;
    switch (workloadType) {
    case ov::WorkloadType::DEFAULT:
        zeWorkloadType = ze_command_queue_workload_type_t::ZE_WORKLOAD_TYPE_DEFAULT;
        break;
    case ov::WorkloadType::EFFICIENT:
        zeWorkloadType = ze_command_queue_workload_type_t::ZE_WORKLOAD_TYPE_BACKGROUND;
        break;
    default:
        OPENVINO_THROW("Unknown value for WorkloadType!");
    }

    for (auto& queue : _command_queues) {
        queue->setWorkloadType(zeWorkloadType);
    }
}

void ZeroExecutor::setArgumentValue(uint32_t argi_, const void* argv_) const {
    zeroUtils::throwOnFail("zeGraphSetArgumentValue", _graph_ddi_table_ext->pfnSetArgumentValue(_graph, argi_, argv_));
}

void ZeroExecutor::mutexLock() const {
    _mutex.lock();
}

void ZeroExecutor::mutexUnlock() const {
    _mutex.unlock();
}

ZeroExecutor::~ZeroExecutor() {
    _logger.debug("~ZeroExecutor() - pfnDestroy _graph ");
    auto result = _graph_ddi_table_ext->pfnDestroy(_graph);
    if (ZE_RESULT_SUCCESS != result) {
        _logger.error("_graph_ddi_table_ext->pfnDestroy failed %#X", uint64_t(result));
    }
}
